/*
  BLE2WiFi_bleScanner for Adafruit Bluefruit nRF52832 Feather
  (c)2022 Forward Computing and Control Pty. Ltd.
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.

  // based on
  *********************************************************************
  This is an example for our nRF52 based Bluefruit LE modules

  Pick one up today in the adafruit shop!

  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  MIT license, check LICENSE for more information
  All text above, and the splash screen below must be included in
  any redistribution
*********************************************************************/

#include <bluefruit.h>
// install the SafeString library via the Arduino Library manager
#include <SafeString.h>
#include "millisDelay.h"
#include "pfodLinkedPointerList.h" // iterable linked list of pointers to objects
#include "LastSeen.h"  // class for storing on linked list

//#define DEBUG

pfodLinkedPointerList<LastSeen> listOfLastSeen;
const size_t MAX_BLE_NAME_SIZE = 32;
const unsigned long BAUD_RATE = 115200ul;

// returns NULL if not found
LastSeen* getLastSeen(SafeString &name) {
  LastSeen *devicePtr = listOfLastSeen.getFirst();
  while (devicePtr) {
    if (name.startsWith(devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = listOfLastSeen.getNext();
  }
  return devicePtr; // null not found
}

const unsigned long SKIP_TIME_MS = 11ul * 1000; // skip adverts for 11sec after first one

void setup() {
  Serial.begin(BAUD_RATE);
  while ( !Serial ) delay(10);   // for nrf52840 with native usb
  SafeString::setOutput(Serial); // for SafeString error msgs
  Serial.println("BLE2WiFi_bleScanner");
  Serial.println("------------------------------------\n");

  // add the names we are interested in NOTE the trailing , so that T_1, does not match T_11,
  LastSeen *devicePtr = new LastSeen("T_0,"); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
  listOfLastSeen.add(devicePtr);
  // add other devices here that you want to scan for
  devicePtr = new LastSeen("T_1,");
  listOfLastSeen.add(devicePtr);
  //  devicePtr = new LastSeen("W_1,");
  //  listOfLastSeen.add(devicePtr);

  // Initialize Bluefruit with maximum connections as Peripheral = 0, Central = 1
  // SRAM usage required by SoftDevice will increase dramatically with number of connections
  Bluefruit.begin(0, 1);
  Bluefruit.setTxPower(4);    // Check bluefruit.h for supported values

  /* Set the device name */
  Bluefruit.setName("Bluefruit52");

  /* Set the LED interval for blinky pattern on BLUE LED */
  Bluefruit.setConnLedInterval(250);

  /* Start Central Scanning
     - Enable auto scan if disconnected
     - Filter out packet with a min rssi
     - Interval = 100 ms, window = 50 ms
     - Use active scan (used to retrieve the optional scan response adv packet)
     - Start(0) = will scan forever since no timeout is given
  */
  Bluefruit.Scanner.setRxCallback(scan_callback);
  Bluefruit.Scanner.restartOnDisconnect(true);
  // Bluefruit.Scanner.filterRssi(-80); // adding a filter here will filter out weak signils from other deivces
  Bluefruit.Scanner.setInterval(160, 80);       // in units of 0.625 ms
  Bluefruit.Scanner.useActiveScan(false);        // Don't Request scan response data
  Bluefruit.Scanner.start(0);                   // 0 = Don't stop scanning after n seconds

  Serial.print("Scanning for ");
  LastSeen *lastSeenPtr = listOfLastSeen.getFirst();
  if (lastSeenPtr == NULL) {
    Serial.println(" Error: no devices specified.");
  } else {
    while (lastSeenPtr != NULL) {
      Serial.print(lastSeenPtr->getDeviceName()); Serial.print(" ");
      lastSeenPtr = listOfLastSeen.getNext();
    }
  }
  Serial.println();
}

/**
   Callback invoked when scanner pick up an advertising data
   @param report Structural advertising data
*/
void scan_callback(ble_gap_evt_adv_report_t* report) {
  char nameBuffer[MAX_BLE_NAME_SIZE + 1]; // for trailing null
  memset(nameBuffer, 0, sizeof(nameBuffer)); // terminate buffer
  cSFA(sfName, nameBuffer); // wrap buffer in SF

  /* Complete Local Name */
  if (Bluefruit.Scanner.parseReportByType(report, BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, (uint8_t*)nameBuffer, sizeof(nameBuffer))) {
    /* Complete Local Name */
  } else if (Bluefruit.Scanner.parseReportByType(report, BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME, (uint8_t*)nameBuffer, sizeof(nameBuffer))) {
    /* Shortened Local Name */
    // receiving ESP will detect truncated data >= 32
  } else {
    // no advertised local name
    Bluefruit.Scanner.resume();
    return;
  }
  sfName.trim();
  // search listOfLastSeen
  LastSeen *devicePtr = getLastSeen(sfName);
  if (!devicePtr) {
    // not found
#ifdef DEBUG
    Serial.print("Not scanning for "); Serial.println(sfName);
#endif
    Bluefruit.Scanner.resume();
    return;
  }
  // else found it check if seen recently
  unsigned long now = millis();
  if ((now - devicePtr->getLastSeen()) > SKIP_TIME_MS) {
    // need to update this
    devicePtr->updateLastSeen(now);
#ifdef DEBUG
    Serial.print("Scan found "); Serial.println(sfName);
#endif
    Serial.println(sfName); // output name found
  } else {
#ifdef DEBUG
    Serial.print("Skip repeat advertising for "); Serial.print(sfName); Serial.println(" seen recently");
#endif
  }

  // For Softdevice v6: after received a report, scanner will be paused
  // We need to call Scanner resume() to continue scanning
  Bluefruit.Scanner.resume();
}

void loop() {
  // do nothing
}
